/* NEC Vr4131 CSI driver
 * CSI: Clocked Serial Interface (SPI)
 *
 * (C) Copyright 2007 Media Lab. <linux@mlb.co.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */

/*#define DEBUG*/
#include <config.h>
#include <common.h>
#include <command.h>
#include <asm/io.h>
#include <spi.h>

#if defined(CONFIG_SPI) && defined(CONFIG_VR41XX_CSI)
#define CSI_BASE 	(KSEG1+0x0f0001a0)
#define CSI_MODE	(volatile u8 *)(CSI_BASE+0x00)
#define  CSI_MODE_CSIE		(1<<7)
#define  CSI_MODE_TRMD		(1<<6)
#define  CSI_MODE_CCL		(1<<5)
#define  CSI_MODE_LSBFIRST	(1<<4)
#define  CSI_MODE_AUTO		(1<<2)
#define  CSI_MODE_CSOT		(1)
#define CSI_CLKSEL	(volatile u8 *)(CSI_BASE+0x01)
#define  CSI_CLKSEL_DAP		(1<<3)
#define  CSI_CLKSEL_CKP		(1<<4)
#define  CSI_CLKSEL_9216K	(1)
#define  CSI_CLKSEL_4608K	(2)
#define  CSI_CLKSEL_2304K	(3)
#define  CSI_CLKSEL_1152k	(4)
#define  CSI_CLKSEL_576K	(5)
#define  CSI_CLKSEL_288K	(6)
#define CSI_SIRB	(volatile u16 *)(CSI_BASE+0x02)
#define CSI_SOTB	(volatile u16 *)(CSI_BASE+0x04)
#define CSI_SIRBE	(volatile u16 *)(CSI_BASE+0x06)
#define CSI_SOTBF	(volatile u16 *)(CSI_BASE+0x08)
#define CSI_SIO		(volatile u16 *)(CSI_BASE+0x0a)
#define CSI_CNT		(volatile u16 *)(CSI_BASE+0x010)  
#define  CSI_CNT_CSIRST		(1<<15)
#define  CSI_CNT_T_TRGEN	(1<<12)
#define  CSI_CNT_T_FIFOF	(1<<11)
#define  CSI_CNT_T_FIFOE	(1<<10)
#define  CSI_CNT_T_P1STP	(1<<9)
#define  CSI_CNT_T_DMAEN	(1<<8)
#define  CSI_CNT_SIO_V		(1<<7)
#define  CSI_CNT_SIRB_V		(1<<6)
#define  CSI_CNT_R_FULLS	(1<<5)
#define  CSI_CNT_R_TRGRE	(1<<4)
#define  CSI_CNT_R_FIFOF	(1<<3)
#define  CSI_CNT_R_FIFOE	(1<<2)
#define  CSI_CNT_R_P1STP	(1<<1)
#define  CSI_CNT_R_DMAEN	(1<<0)
#define CSI_INT		(volatile u16 *)(CSI_BASE+0x012)
#define  CSI_INT_END		(1<<15)
#define  CSI_INT_T_P2STP	(1<<10)
#define  CSI_INT_T_P1STP	(1<<9)
#define  CSI_INT_T_EMP		(1<<8)
#define  CSI_INT_R_P2STP	(1<<2)
#define  CSI_INT_R_P1STP	(1<<1)
#define  CSI_INT_R_OVER		(1<<0)
#define  CSI_INT_MASK		(0x8707)
#define CSI_IFIFOV	(volatile u16 *)(CSI_BASE+0x014)
#define CSI_OFIFOV	(volatile u16 *)(CSI_BASE+0x016)
#define CSI_IFIFO	(volatile u16 *)(CSI_BASE+0x018)
#define CSI_OFIFO	(volatile u16 *)(CSI_BASE+0x01a)
#define CSI_FIFOTRG	(volatile u16 *)(CSI_BASE+0x01c)
#define  CSI_FIFOTRG_T_TRG(n)	((n)<<8)
#define  CSI_FIFOTRG_R_TRG(n)	(n)

#define CMU_CLK_MSK		(KSEG1+0x0F000060) /* clock mask */
#define CMU_CLK_MSK_SCSI	(1<<12)
#define CMU_CLK_MSK_CSI		(1<<6)


/*#define DATA_LENGTH_16		1*/


#define csi_writew( reg, val )	(*(reg) = (val))
#define csi_writeb( reg, val )	(*(reg) = (val))
#define csi_readw( reg )	(*(reg))
#define csi_busyloop()		while ( csi_readw( CSI_MODE ) & CSI_MODE_CSOT )

void spi_init (void)
{
	debug ( "%s at %08X\n", __FUNCTION__, CSI_BASE );
	/* enable clock for CSI Unit */
	*((volatile u16 *)(CMU_CLK_MSK)) |= CMU_CLK_MSK_CSI|CMU_CLK_MSK_SCSI;
	csi_writew( CSI_CNT, CSI_CNT_CSIRST ); 
}

int spi_xfer(spi_chipsel_type chipsel, int bitlen, uchar *out, uchar *in)
{
	int len, i;
	u16 mode;
#ifdef DATA_LENGTH_16
	ushort *dout = (ushort *)out;
	ushort *din  = (ushort *)in;
#else
	uchar *dout = out;
	uchar *din  = in;
#endif
	debug ( "%s(%p,%d,%p,%p)\n", __FUNCTION__, chipsel, bitlen, dout, din );

#ifdef DATA_LENGTH_16
	len = (bitlen + 15) / 16;
	mode = CSI_MODE_TRMD|CSI_MODE_CCL;
#else
	len = (bitlen + 7) / 8;
	mode = CSI_MODE_TRMD;
#endif

	/* deassert CSI reset and mode set */
	csi_writew( CSI_CNT, 0 ); 

	/* 8bit, MSB first */
	csi_writeb( CSI_MODE, mode );
	csi_writeb( CSI_CLKSEL, CSI_CLKSEL_CKP|CSI_CLKSEL_576K ); 
	csi_writew( CSI_MODE, mode | CSI_MODE_CSIE );

	if(chipsel != NULL) {
		(*chipsel)(1);	/* select the target chip */
	}

#ifdef DEBUG
	debug ( "out(%d):", len );
	for ( i = 0; i < len; i++ )
	  debug ( "%02X ", dout[i] );
	debug ( "\n" );
#endif	

	while ( len-- ) {
		csi_writew( CSI_MODE, mode | CSI_MODE_CSIE );
		csi_writew( CSI_SOTB, *(dout++) );
		csi_busyloop();
		*(din++) = csi_readw( CSI_SIO );
		csi_writew( CSI_INT, csi_readw(CSI_INT) );
		csi_writew( CSI_MODE, mode );
	}

	if(chipsel != NULL) {
		(*chipsel)(0);	/* deselect the target chip */
	}

#ifdef DEBUG
	debug ( "in(%d):", ((uchar*)din) - in );
	for ( i = 0; &in[i] < (uchar*)din; i++ )
	  debug ( "%02X ", in[i] );
	debug ( "\n" );
#endif	
	csi_writew( CSI_CNT, CSI_CNT_CSIRST ); 
	return(0);
}

#endif /* defined(CONFIG_SPI) && defined(CONFIG_VR4131_CSI) */
