/*
 * ifnic.c network driver for CQ Publishing's Blanca base board.
 * this device is soft core, build on FPGA.
 *
 * (C) Copyright 2007 Media Lab. <linux@mlb.co.jp>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation
 */
/* #define DEBUG	1 */
#include <common.h>
#include <malloc.h>
#include <net.h>
#include <asm/io.h>
#include <pci.h>
#include <miiphy.h>

#include "ifnic.h"
#define PHY_ADDR	6

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) && \
	(defined(CONFIG_IFNIC)||defined(CONFIG_IFNIC_PCI))

#ifdef DEBUG
static int verbose = 0;
#define debug_verbose_start	verbose++
#define debug_verbose_stop	verbose--
#define debug_verbose(fmt, args...)	if(verbose>0) printf(fmt ,##args)
#else
#define debug_verbose_start
#define debug_verbose_stop
#define debug_verbose(fmt, args...)
#endif

static inline void if_writel(u_long addr, u32 data)
{
	debug_verbose( "%s( %08X, %08X)\n", __FUNCTION__, addr, cpu_to_le32(data) );
	*(volatile u32 *)addr = cpu_to_le32(data);
}

static inline u32 if_readl(u_long addr)
{
	debug_verbose( "%s( %08X ) = %08X\n", __FUNCTION__,
		       addr, *(volatile u32 *)addr );
	return le32_to_cpu(*(volatile u32 *)addr);
}

static void wait_bitclear(u_long base, u_long bit)
{
	while( if_readl(base) & bit )
		;
}

#define phy_wait(dev)	wait_bitclear((dev->iobase+IFNIC_MIIPHY), MIIPHY_BUSY)
#define rom_wait(base)	wait_bitclear((base+IFNIC_ROM), ROM_BUSY)
#define tx_wait(base)	wait_bitclear((base+IFNIC_TXBUF_STATUS), TXBUF_BUSY)

static int phy_read(struct eth_device *dev, int phyaddr, int reg )
{
	phy_wait(dev);

	/* Issue the MII 'read' command. */
	if_writel(dev->iobase + IFNIC_MIIPHY,
		  MIIPHY_BUSY|MIIPHY_READ|MIIPHY_ADDR(phyaddr)|MIIPHY_REG(reg));

	/* Wait for the command to complete. */
	phy_wait(dev);
	return MIIPHY_READDATA(if_readl(dev->iobase+IFNIC_MIIPHYDATA));
}

static int phy_write(struct eth_device *dev, int phyaddr, int reg, unsigned short const value )
{
	phy_wait(dev);

	/* set write value */
	if_writel(dev->iobase + IFNIC_MIIPHYDATA, value);

	/* Issue the MII 'write' command. */
	if_writel(dev->iobase + IFNIC_MIIPHY, MIIPHY_BUSY|MIIPHY_ADDR(phyaddr)|MIIPHY_REG(reg));

	/* Wait for the command to complete. */
	phy_wait(dev);
	return 0;
}

static int rom_read(u_long base, u_short offset)
{
	rom_wait(base);

	/* Issue the ROM 'read' command. */
	if_writel(base + IFNIC_ROM, ROM_BUSY|ROM_READ|ROM_ADDR(offset));

	/* Wait for the command to complete. */
	rom_wait(base);

	return ROM_READDATA( if_readl( base + IFNIC_ROM ) );
}

#if defined(CONFIG_MII) || (CONFIG_COMMANDS & CFG_CMD_MII)
/**
 * Read a 16-bit value from an MII register.
 */
static int ifnic_miiphy_read(char * const devname, unsigned char const addr,
		unsigned char const reg, unsigned short * const value)
{
	struct eth_device *dev = eth_get_dev_by_name(devname);
	int ret;

	/* Parameter checks */
	if (devname == NULL || dev == NULL) {
		printf(("NULL dev"));
		return -1;
	}

	debug_verbose_start;
	ret = phy_read(dev, addr, reg );
	debug_verbose_stop;

	if ( ret < 0 ) {
		return -1;
	}
	*value = ret;
	return 0;
}


/**
 * Write a 16-bit value to an MII register.
 */
static int ifnic_miiphy_write(char * const devname, unsigned char const addr,
		unsigned char const reg, unsigned short const value)
{
	struct eth_device *dev = eth_get_dev_by_name(devname);

	/* Parameter checks */
	if (devname == NULL || dev == NULL) {
		printf(("NULL dev"));
		return -1;
	}

	debug_verbose_start;
	if ( phy_write(dev, addr, reg, value ) ) {
	  debug_verbose_stop;
		return -1;
	}
	debug_verbose_stop;
	return 0;
}

#endif

static int ifnic_linkcheck(struct eth_device* dev)
{
	u16 bmcr;
	u32 link = if_readl( dev->iobase + IFNIC_MIISTATUS );

	if ( link & (DUPLEX_CHANGE|SPEED_CHANGE|LINK_CHANGE) ) {
		/* clear flags */
		if_writel( dev->iobase + IFNIC_MIISTATUS, link );
		if (!(link & LINK_STATUS)) {
			printf( "link down, cable problem ?\n" );
			return -1;
		}
		printf( "Link status:" );
		bmcr = phy_read(dev, PHY_ADDR, PHY_BMCR ) & PHY_BMCR_AUTON;
		if (link & SPEED_STATUS){
			bmcr |= PHY_BMCR_100_MBPS;
			printf( "100MBps" );
		} else {
			bmcr |= PHY_BMCR_10_MBPS;
			printf( "10MBps" );
		}
		if (link & DUPLEX_STATUS) {
			printf( "FDX" );
			bmcr |= PHY_BMCR_DPLX;
		} else {
			printf( "HDX" );
		}
		phy_write( dev, PHY_ADDR, PHY_BMCR, bmcr );
		printf( "\n" );
		return 1;
	}
	return 0;
}

static int ifnic_init(struct eth_device* dev, bd_t* bis)
{
	int i;
	u32 link;
	debug( "%s\n",__FUNCTION__ );
	
	/* set Mac addr */
	debug_verbose_start;
	for (i = 0; i < 6; i++) {
		if_writel( dev->iobase + IFNIC_MAC_ADDR,
			   MAC_WRITE|MAC_ADDR(i)|MAC_DATA(dev->enetaddr[i]));
	}
	debug_verbose_stop;

	link = if_readl( dev->iobase + IFNIC_MIISTATUS );
	if (!(link & LINK_STATUS)) {
		/* clear flags */
		if_writel( dev->iobase + IFNIC_MIISTATUS, link );
		printf( "link is down, cable problem ?\n" );
	}else
		ifnic_linkcheck(dev);
	if_writel( dev->iobase + IFNIC_RXCFG, 0 /*|NO_BROADCAST*/ );
	if_writel( dev->iobase + IFNIC_TXCFG, 0 );
	return 1;
}

static int ifnic_send(struct eth_device* dev, volatile void* packet, int length)
{
	debug( "%s(%d)\n",__FUNCTION__, length );

	/* wait for last transmisson to complete  */
	tx_wait(dev->iobase);

	/* Write to send buffer */
	memcpy( (void*)(dev->iobase + IFNIC_TXBUF), (void*)packet, length );

	if ( length < 64 ) length = 64; /* Smallest possible ethernet frame size*/

	/* start transmisson */
	if_writel(dev->iobase + IFNIC_TXBUF_STATUS, (TXBUF_BUSY | (length & TXBUF_SIZE_MASK)));

	return 1;/* OK */
}

static int ifnic_recv(struct eth_device* dev)
{
	int size = if_readl(dev->iobase+IFNIC_RXBUF_STATUS);

	if ( size & (RXBUF_OF|RXBUF_PHYERR|RXBUF_FERR2|RXBUF_FERR1|RXBUF_FERR0|RXBUF_CRCERR) ){
		printf( "Rx Err %08X\n", size );
		if_writel(dev->iobase+IFNIC_RXBUF_STATUS,RXBUF_FREE);
		return 0;
	}
	if ( ! (size & RXBUF_VALID) ) {
		ifnic_linkcheck(dev);
		return 0;
	}
	debug( "%s state=0x%08X\n",__FUNCTION__, size );

	size &= RXBUF_SIZE_MASK;

	NetReceive((volatile uchar *)(dev->iobase + IFNIC_RXBUF), size );
	if_writel(dev->iobase+IFNIC_RXBUF_STATUS, RXBUF_FREE);
	return 1;
}

static void ifnic_halt(struct eth_device* dev)
{
	debug( "%s\n",__FUNCTION__ );
	if_writel( dev->iobase + IFNIC_RXCFG, RXRST );
	if_writel( dev->iobase + IFNIC_TXCFG, TXRST );
}

static int ifnic_hw_initialize(struct eth_device *dev)
{
	int i;
	u32 id;

	/* sanity check */
	id = if_readl(dev->iobase + IFNIC_DEVID);
	if (id != IFNIC_DEVID_VALUE) {
		printf("%s: Error, ID not match.\n", dev->name );
		return -1;
	}

	id = if_readl(dev->iobase + IFNIC_REV);
	for (i = 0; i < 6; i++) {
		dev->enetaddr[i] = rom_read( dev->iobase, IFNIC_MAC_ROMADDR + i );
	}
	debug("%s rev %02X.%02X at %08X: %02x:%02x:%02x:%02x:%02x:%02x\n",
	      dev->name, rev>>16, rev&0xFF, dev->iobase, 
	      dev->enetaddr[0], dev->enetaddr[1], dev->enetaddr[2],
	      dev->enetaddr[3], dev->enetaddr[4], dev->enetaddr[5]);

	if_writel( dev->iobase + IFNIC_INTMASK, 0);
	if_writel( dev->iobase + IFNIC_INTSTAT, 0xFFFFFFFF);
	if_writel( dev->iobase + IFNIC_MIISTATUS, 0xFFFFFFFF);
	if_writel( dev->iobase + IFNIC_RXCFG, RXRST);
	if_writel( dev->iobase + IFNIC_TXCFG, TXRST);
	if_writel( dev->iobase + IFNIC_RXBUF_STATUS, RXBUF_FREE);
	if_writel( dev->iobase + IFNIC_STAT_LOST, 0);
	if_writel( dev->iobase + IFNIC_STAT_PHYERR, 0);
	if_writel( dev->iobase + IFNIC_STAT_FRAMEERR, 0);
	if_writel( dev->iobase + IFNIC_STAT_CRCERR, 0);
	if_writel( dev->iobase + IFNIC_STAT_COLLISION, 0);

	/* reset PHY */
	phy_write(dev, PHY_ADDR, PHY_BMCR, PHY_BMCR_RESET|PHY_BMCR_AUTON|PHY_BMCR_RST_NEG );

	dev->init = ifnic_init;
	dev->send = ifnic_send;
	dev->recv = ifnic_recv;
	dev->halt = ifnic_halt;
	return 0;
}

/*
 * Probe for the device
 */
#ifdef CONFIG_IFNIC_PCI
extern int ifnic_initialize(bd_t * bis)
{
	struct eth_device *dev;
	pci_dev_t devno;
	int cardno = 0, i;
	u32 iobase;
	u32 cmd, rev;

	debug( "%s\n",__FUNCTION__ );
	/* Find PCI device(s) */
	while ((devno = pci_find_device(PCI_VENDOR_ID_KURUSUGAWA_ELE,
					PCI_DEVICE_ID_BLANCA_NIC, cardno))>=0)
	{
		pci_read_config_dword(devno, PCI_BASE_ADDRESS_0, &iobase);

		/* Enable Memory space */
		pci_write_config_dword(devno, PCI_COMMAND, PCI_COMMAND_MEMORY );

		/* Check if memory accesses is enabled. */
		pci_read_config_dword(devno, PCI_COMMAND, &cmd);
		if (!(cmd & PCI_COMMAND_MEMORY)) {
			printf("ifnic#%d: Error Can not enable MEM space.\n", cardno );
			continue;
		}
		dev = (struct eth_device *) malloc(sizeof (*dev));
		dev->priv = (void *)devno;
		dev->iobase = pci_mem_to_phys(devno, iobase);
		sprintf(dev->name, "ifnic#%d", cardno);

		if (ifnic_hw_initialize(dev) < 0 ) {
			free(dev);
			continue;
		}
		eth_register(dev);
#if defined(CONFIG_MII) || (CONFIG_COMMANDS & CFG_CMD_MII)
		miiphy_register(dev->name, ifnic_miiphy_read, ifnic_miiphy_write );
#endif
		cardno++;
	}
	if ( cardno )
		return 0;
	return -1;
}

#else /* CONFIG_IFNIC */
extern int ifnic_initialize(bd_t * bis)
{
	struct eth_device *dev;
	dev = (struct eth_device *) malloc(sizeof (*dev));
	dev->iobase = CONFIG_IFNIC;
	strcpy(dev->name, "ifnic");

	if (ifnic_hw_initialize(dev) < 0 ) {
		free(dev);
		return -1;
	}
	eth_register(dev);
#if defined(CONFIG_MII) || (CONFIG_COMMANDS & CFG_CMD_MII)
	miiphy_register(dev->name, ifnic_miiphy_read, ifnic_miiphy_write );
#endif
	return 0;
}
#endif /* CONFIG_IFNIC */

#endif /* (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) && \
	(defined(CONFIG_IFNIC)||defined(CONFIG_IFNIC_PCI)) */
